/*
   Gzilla's IO Queue Processor
   Randall Maas
   1999
*/

#include <malloc.h>
#include <stdio.h>
#include <sys/wait.h>
#include <signal.h>
#include <search.h>
#include <sys/socket.h>
#include "IOQ.h"
#include "LDL.h"
#include "GzIO.h"
#include "../gzillacache.h"
//#include <gtk/gtk.h>
#include "../gzilla.h"

static volatile unsigned Flags=0;
#define GZ_RUNIO (1)
#define GZ_IDLELOOP (2)
struct qelem Gz_SubmitQ={&Gz_SubmitQ,&Gz_SubmitQ},Gz_ComplQ[IO_NPRIO],Gz_FreeQ;
int NChildren=0;

void GzIO_abort(int FD)
{
   __IO_abort((__IO_req_t*)&Gz_SubmitQ, (__IO_req_t*)&Gz_FreeQ, FD);
  close(FD);
/* BUg: already completed ones will get resubmitted! */
}

static int GZBS_Idle_Function(void*);
static void SIG_handler(int sig)
{
   int X;
   Flags |= GZ_RUNIO;
   while (waitpid(-1,&X,WNOHANG)>0) {NChildren--;}
   if (!(Flags & GZ_IDLELOOP))
     {
       Flags |= GZ_IDLELOOP; 
       Gzilla_idle_add(GZBS_Idle_Function, NULL);
     }
}

void GzIO_init(void)
{
   struct sigaction Act1;
   int I;
   LDL_init(&Gz_FreeQ);
   for(I=0;I<IO_NPRIO; I++)
    {
       LDL_init(&Gz_ComplQ[I]);
    }
   FD_ZERO(&Gz_FDs);
   Act1.sa_handler=SIG_handler;
   sigemptyset(&Act1.sa_mask);
   sigaddset(&Act1.sa_mask,SIGIO);
   sigaddset(&Act1.sa_mask,SIGCHLD);
   sigaddset(&Act1.sa_mask,SIGPIPE);
   Act1.sa_flags=SA_RESTART;
   sigaction(SIGIO, &Act1, NULL);
   sigaction(SIGCHLD, &Act1, NULL);
   sigaction(SIGPIPE, &Act1, NULL);
}

__Gzilla_IO_t* GzIO_new(int FD, void* Data, size_t Size)
{
   __Gzilla_IO_t* IPtr;
   if (!LDL_empty(&Gz_FreeQ))
     {
	IPtr = (__Gzilla_IO_t*) Gz_FreeQ.q_forw;
	remque((struct qelem*) IPtr);
     }
    else
     {
	IPtr = (__Gzilla_IO_t*) malloc(sizeof(*IPtr));
	IPtr->IO.Count=1;
	IPtr->IO.IOReq= &IPtr->IOVec;
     }
   IPtr->Cache_FD=-1; // There is no file to write the data too
   IPtr->IO.Op = 0;   // A read... always a read
   IPtr->IO.FD = FD;
   IPtr->IOVec.iov_base = Data;
   IPtr->IOVec.iov_len  = Size;
   return IPtr;
}

void GzIO_Submit(__Gzilla_IO_t* IPtr)
{
   LDL_append((struct qelem*) IPtr, &Gz_SubmitQ);
   FD_SET(IPtr->IO.FD,&Gz_FDs);
   Flags |= GZ_RUNIO;
   if (!(Flags & GZ_IDLELOOP))
     {
        Flags |= GZ_IDLELOOP; 
        Gzilla_idle_add(GZBS_Idle_Function,NULL);
     }
}

static int IO_NextRead(int FD, __CacheFile_t* DPtr)
{
   /*
      Description
       This creates the next read on the IO channel into our cache memory
       buffers.

      Return Value
       0 == Ok, otherwise badness
   */
   size_t Transfer_Size; //Bug: use fd block size
   __Gzilla_IO_t* IPtr;

   Transfer_Size = DPtr->Block_Size;
   if (!DPtr->Total_Size)
     {
        void *NewData;
        /* Allocate more memory chunks */
        if (!(NewData = realloc((void*)DPtr->Data, DPtr->Size+Transfer_Size)))
	  {
	     /* Bad things have happened... we didn't get any more space */
	     return -1;
	  }

	/* Update our pointer to the memory... */
	DPtr->Data = NewData;
     }
   else
     {
        /* Check to see if we are done...*/
        if (DPtr->Size >= DPtr->Total_Size) return -1;

	/* Figure out how to top the cache memory buffer off */
        if (DPtr->Block_Size+DPtr->Size > DPtr->Total_Size)
	 Transfer_Size = DPtr->Total_Size-DPtr->Size;
     }

   /* Indicate where our next read is going to in memory... */
   IPtr = GzIO_new(FD, ((char*)DPtr->Data) + DPtr->Size, Transfer_Size);
   if (!IPtr) return -1;
   GzIO_Submit(IPtr);

   return 0;
}

static void IO_complete(__Gzilla_IO_t* IPtr)
{
   /*
      Description
       This routine handles a successfully completed read.
   */
   __CacheFile_t* DPtr = FD2Cache(IPtr->IO.FD);
   __IOCallback_t Call = _FD2Ptr(IPtr->IO.FD).Call;
   void* Data = _FD2Ptr(IPtr->IO.FD).Data;
   int FreeMe=0;

   /* First update our cache memory stuff */
   /* Update the size of valid memory that we have */
   DPtr->Size+= IPtr->IO.Status;

   /* Next update our client */
   if (Call) Call(0, Data,DPtr);

   /* Third, create a new read */
   if (!IPtr->IO.Op)
      if (IO_NextRead(IPtr->IO.FD, DPtr))
	{
	   /* Okay, something bad has happened, or we should otherwise close
	      this puppy */
	   shutdown(IPtr->IO.FD,2);
	   close(IPtr->IO.FD);
	   if (Call) Call(1,Data,DPtr);
	   FreeMe=1;
	}
   
   /* Is this also supposed to go to the cache? */
   if (IPtr->Cache_FD < 0)
     {
       /* Queue up some more stuff read .... */
       if (FreeMe && (DPtr->Flags & GZ_AUTOFREE))
	 {
	    LDL_append(DPtr,&CacheFree);
	    free((void*)DPtr->Data);
	 }
       return;
     }

   /* It /is/ also supposed to go to the cache... */
   IPtr->IO.FD = IPtr->Cache_FD; /* Change the file descriptor */
   IPtr->IO.Op = 1;		 /* Change the operation to write */
   IPtr->Cache_FD = -1;

   /* Now queue this IO up for more fun... */
   remque(IPtr);
   GzIO_Submit(IPtr);
}

static void GZBS_Clean_Queue(struct qelem* CQ)
{
   /* 
      Description
       This routine dequeues all of the completed IO -- aborted and good.
       It routes the data to the proper place in the world.
    */
   __IO_req_t* _IPtr,* _JPtr;
# define IPtr ((__Gzilla_IO_t*)_IPtr)
   __IOCallback_t Call;
   void* Data;

   LDL_foreach(_IPtr, (__IO_req_t*)CQ)
    {
    PAST_GO:
       Call = _FD2Ptr(IPtr->IO.FD).Call;
       Data = _FD2Ptr(IPtr->IO.FD).Data;

       /* See if the thing is a dead one, and needs to be closed */
       if (IPtr->IO.Status <1)
         {
	    /* Pass the notification on if this isn't an IO for writing to
		the cache and there isn't another FD to chain to... */
	    __CacheFile_t* FPtr = FD2Cache(IPtr->IO.FD);
	    if (!IPtr->IO.Op)
	      {
	 	 if (Call) Call(1, Data,FPtr);
		 /* Close the cache channel too */
		 if (IPtr->Cache_FD)
		   {
		      close(IPtr->Cache_FD);
		      FD_CLR(IPtr->IO.FD, &Gz_FDs);
		   }
		 if (FPtr->Flags & GZ_AUTOFREE)
		   {
		      LDL_append(FPtr, &CacheFree);
		      free((void*)FPtr->Data);
		   }
	      }
	    FD_CLR(IPtr->IO.FD, &Gz_FDs);
	    shutdown(IPtr->IO.FD,2);
	    close(IPtr->IO.FD);
	    _FD2Ptr(IPtr->IO.FD).Data = NULL;
	    _FD2Ptr(IPtr->IO.FD).Call = NULL;
	    continue;
	 }

       /* Complete the IO. 
	  Check to see if it is a completion of a write to cache
	  (we do it here -- instead of in IO_complete) to speed processing..
	*/
       if (IPtr->IO.Op) continue;

       /* Replace the foreach's next function... */
       _JPtr = _IPtr->q_forw;
       IO_complete(IPtr);

       if (!_JPtr || _JPtr == (typeof(_JPtr))CQ) break;
       _IPtr = _JPtr;
       goto PAST_GO;
    }
}

static int GZBS_Idle_Function(void* Ptr)
{
   /* 
      Description
       This routine dequeues all of the completed IO -- aborted and good.
       It routes the data to the proper place in the world.
    */
   int X;

   while (waitpid(-1,&X,WNOHANG)>0) {NChildren--;}
   /* See if we need to run the IO handler... */
   if (!(Flags & GZ_RUNIO))
     {
        Flags &= ~GZ_IDLELOOP;
	return 0;
     }

   Flags &= ~GZ_RUNIO;
   if (__IOQ_run((__IO_req_t*)&Gz_SubmitQ,(__IO_req_t**)&Gz_ComplQ,
		 (__IO_req_t**)&Gz_ComplQ, &Gz_FDs))
     {
        int I;

	for (I = 0; I < IO_NPRIO; I++)
	 {
	    GZBS_Clean_Queue(&Gz_ComplQ[I]);
	    /* Free all of the finished IO */
	    LDL_cat(&Gz_FreeQ, &Gz_ComplQ[I]);
	    LDL_init(&Gz_ComplQ[I]);
	 }
     }

   if (!(Flags & GZ_RUNIO))
     {
         Flags &= ~GZ_IDLELOOP;
	 return 0;
     }
   return 1; //Keep on running....
}
