/*

 gzilla

 Copyright 1997 Raph Levien <raph@acm.org>

 This code is free for commercial and non-commercial use,
 modification, and redistribution, as long as the source code release,
 startup screen, or product packaging includes this copyright notice.

 */


/* todo: make error messages more descriptive. */

#include <unistd.h>

#include <errno.h>            /* for errno */
#include <string.h>           /* for strstr */
#include <malloc.h>
#include <stdlib.h>
#include <search.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/wait.h>
#include "../IO/GzIO.h"
#include "URL_protos.h"

#include "gzillaurl.h"
#include "../gzillasocket.h"
#include "../gzillaweb.h"
#include "../config.h"

typedef struct __Info_t
{
   struct __Info_t *q_forw,*prev;
   char* Query;
} __Info_t;

static __Info_t IQue={NULL,NULL};

/* The callback for socket setup. This function gets called whenever
   the socket setup completes. Install the input handler for writing
   the query string to the socket. */

static void gzilla_http_callback (int Op, int FD_Sock, void* VPtr)
{
   /* Send the header stuff of to the Type channel, this is somewhat expensive
       since we are doing a lot of kernel calls to do this safely
     */
   __Info_t* IPtr=(__Info_t*) VPtr;

   /* This could take a while to get the query /back/... So we need to do
	something.. Alas, not much can be done right yet. */
   if (Op)
     {
       if (IPtr->Query)
	 {
	   free(IPtr->Query);
	   IPtr->Query=NULL;
	 }

        // Free this info node
        insque(IPtr, &IQue);
	close(FD_Sock);
	return;
     }

   /* Send the query out... */
   if (IPtr->Query)
     {
	size_t Size=strlen(IPtr->Query);
	// Make this a background IO
	__Gzilla_IO_t* IOPtr = GzIO_new(FD_Sock, IPtr->Query, Size);	
	IOPtr->IO.Op=1;
	GzIO_Submit(IOPtr);
	//Bug IPtr is never freed!
     }
   gzilla_cache_tmp(FD_Sock);
   FD_bkgnd(FD_Sock);
}

/* Create a new http connection for URL url, and asynchronously
   feed the bytes that come back to bytesink. */
int gzilla_http_get (const char *url, void* Data)
{
  char hostname[256];
  int port, Ret;
  char *tail;
  char query[2048];
  __Info_t* IPtr;

  port = 80;

  /* hacked-in support for proxies, inspired by Olivier Aubert */
  if (HTTP_Proxy && !(No_Proxy && strstr (url, No_Proxy) != NULL))
    {
      gzilla_url_parse (HTTP_Proxy, hostname, sizeof(hostname), &port);
      tail = (char *)url;
    }
  else
    tail = gzilla_url_parse ((char *)url, hostname, sizeof(hostname), &port);

  if (!tail) return -1;

  sprintf (query,
	       "GET %s HTTP/1.0\r\n"
	       "User-Agent: gzilla %s\r\n"
	       "Host: %s:%d\r\n"
	       "\r\n", tail, VERSION, hostname, port);
  if (IQue.q_forw)
    {
       IPtr=IQue.q_forw;
       remque(IPtr);
    }
   else
    {
       if (!(IPtr=malloc(sizeof(*IPtr)))) return -1;
    }
  IPtr->Query = strdup (query);

   /* Setup a socket, in a Gzilla friendly fashion... because it is remote,
      we only will have information about the name.  We will get a call back
      later indicating what we should do about that.
    */
   Ret = gzilla_socket_new (hostname, port, gzilla_http_callback, (void*)IPtr);

   //Bug: if socket fails, no way to recover IPtr
   if (Ret < 0)
     {
        insque((struct qelem*)IPtr,(struct qelem*)&IQue);
	free(IPtr->Query);
        fprintf(stderr, "couldn;t get new socket\n");
	return -1;}

   _FD2Ptr(Ret).Call =gzilla_web_callback;
   _FD2Ptr(Ret).Data = Data;
   return Ret;
}
